
__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


# namespace flags enum:
OVERWRITE      =    1
RECURSIVE      =    2
DEREFERENCE    =    4
CREATE         =    8
EXCLUSIVE      =   16
LOCK           =   32
CREATE_PARENTS =   64
# TRUNCATE     =  128 # reserved
# APPEND       =  256 # reserved
# READ         =  512 # reserved
# WRITE        = 1024 # reserved
# READ_WRITE   = 1536 # reserved
# BINARY       = 2048 # reserved




