
__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from radical.saga.adaptors.cpi.replica.logical_file      import LogicalFile
from radical.saga.adaptors.cpi.replica.logical_directory import LogicalDirectory

